#include <stdio.h>
#include <stdint.h>
#include <fcntl.h>
#include "dosfs.h"
#include "FdSerial.h"
#include "propeller.h"

#define SD_INIT_CMD             0x0d    // %011_01
#define SD_READ_CMD             0x11    // %100_01
#define SD_WRITE_CMD            0x15    // %101_01

#ifndef TRUE
#define TRUE    1
#define FALSE   0
#endif

#define MAX_FILES   4
#define BASE_FD     3

typedef struct {
    FILEINFO file;
    int inUse;
} File;

static VOLINFO vinfo;
static File Files[MAX_FILES];
static char scratch[SECTOR_SIZE];
static int mounted = FALSE;

void _initIO(void)
{
    FdSerial_start(serial_data, 31, 30, 0, 115200);
    MountFS();
}

int inbyte(void)
{
    return FdSerial_rx(serial_data);
}

void outbyte(int ch)  
{
    FdSerial_tx(serial_data, ch);
}

static int MountFS(void)
{
    uint8_t active, ptype;
    uint32_t start, psize, err;
    int i;

    iprintf("GetPtnStart:\r\n");
    start = DFS_GetPtnStart(0, scratch, 0, &active, &ptype, &psize);
    if (start == 0xffffffff)
        return FALSE;

    iprintf(" -->%ld\r\n", start);
    iprintf("active %d, ptype %d, psize %lx, start %lx\r\n", active, ptype, psize, start);

    iprintf("GetVolInfo:\r\n");
    err = DFS_GetVolInfo(0, scratch, start, &vinfo);
    iprintf(" -->%ld\r\n", err);
    if (err)
        return FALSE;

	iprintf("Volume label '%-11.11s'\r\n", vinfo.label);
	iprintf("%d sector/s per cluster, %d reserved sector/s, volume total %ld sectors.\r\n", vinfo.secperclus, vinfo.reservedsecs, vinfo.numsecs);
	iprintf("%ld sectors per FAT, first FAT at sector #%ld, root dir at #%ld.\r\n",vinfo.secperfat,vinfo.fat1,vinfo.rootdir);
	iprintf("(For FAT32, the root dir is a CLUSTER number, FAT12/16 it is a SECTOR number)\r\n");
	iprintf("%d root dir entries, data area commences at sector #%ld.\r\n",vinfo.rootentries,vinfo.dataarea);
	iprintf("%ld clusters (%ld bytes) in data area, filesystem IDd as ", vinfo.numclusters, vinfo.numclusters * vinfo.secperclus * SECTOR_SIZE);
	if (vinfo.filesystem == FAT12)
		iprintf("FAT12.\r\n");
	else if (vinfo.filesystem == FAT16)
		iprintf("FAT16.\r\n");
	else if (vinfo.filesystem == FAT32)
		iprintf("FAT32.\r\n");
	else {
		iprintf("[unknown]\r\n");
        return FALSE;
    }

    for (i = 0; i < MAX_FILES; ++i)
        Files[i].inUse = FALSE;

    return TRUE;
}

int write(int fd, char *buf, int nbytes)
{
    int i = fd - BASE_FD;
    uint32_t count;
    if (fd == 1) {
	  for (i = 0; i < nbytes; i++) {
	    if (*(buf + i) == '\n')
	      outbyte ('\r');
	    outbyte (*(buf + i));
	  }
	  return (nbytes);
    }
    if (i < 0 || i >= MAX_FILES)
        return -1;
    if (!Files[i].inUse)
        return -1;
    DFS_WriteFile(&Files[i].file, scratch, buf, &count, nbytes);
    return count > 0 ? count : EOF;
}

int read(int fd, char *buf, int nbytes)
{
    int i = fd - BASE_FD;
    uint32_t count;
    if (fd == 0) {
	  int i = 0;
	  for (i = 0; i < nbytes; i++) {
		char t=inbyte();
		if ((t!='\r')&&(t!='\n')) {
		    *(buf + i) = t;
		    outbyte(t); // echo
		}
        else {
	    	// terminate the line in the way expected by the libraries
		    *(buf + i) = '\n';
	    	i++;
		    outbyte('\r');
		    outbyte('\n');
	        break;
	    }
	  }
	  return (i);
    }
    if (i < 0 || i >= MAX_FILES)
        return -1;
    if (!Files[i].inUse)
        return -1;
    DFS_ReadFile(&Files[i].file, scratch, buf, &count, nbytes);
    return count > 0 ? count : EOF;
}

int open(const char *name, int flags, ...)
{
    int dfs_mode, i;
    switch (flags & O_ACCMODE) {
    case O_RDONLY:
        dfs_mode = DFS_READ;
        break;
    case O_WRONLY:
        dfs_mode = DFS_WRITE;
        break;
    default:
        return -1;
    }
    for (i = 0; i < MAX_FILES; ++i)
        if (!Files[i].inUse)
            break;
    if (i >= MAX_FILES)
        return -1;
    if (DFS_OpenFile(&vinfo, (char *)name, dfs_mode, scratch, &Files[i].file) != DFS_OK)
        return -1;
    Files[i].inUse = TRUE;
    return BASE_FD + i;
}

int close(int fd)
{
    int i = fd - BASE_FD;
    if (i < 0 || i >= MAX_FILES)
        return -1;
    if (!Files[i].inUse)
        return -1;
    Files[i].inUse = FALSE;
    return 0;
}

off_t lseek(int fd, off_t offset, int whence)
{
    return -1;
}

static uint32_t Swap32(uint32_t value)
{
    return  (value >> 24)
         | ((value >>  8) & 0x0000ff00)
         | ((value <<  8) & 0x00ff0000)
         |  (value << 24);
}

static void CopyAndSwapBytes(uint8_t *dst, uint8_t *src, int cnt)
{
    uint32_t *ldst = (uint32_t *)dst;
    uint32_t *lsrc = (uint32_t *)src;
    while ((cnt -= sizeof(uint32_t)) >= 0)
        *ldst++ = Swap32(*lsrc++);
}

uint32_t DFS_ReadSector(uint8_t unit, uint8_t *buffer, uint32_t sector, uint32_t count)
{
    while (count > 0) {
        vm_params[0] = scratch_buffer_off;
        vm_params[1] = sector_size;
        vm_params[2] = sector;
        if (_syscall(NULL, SYS_mboxcmd, vm_mbox_off, SD_READ_CMD | (vm_params_off << 8)) != 0)
            return -1;
        CopyAndSwapBytes(buffer, scratch_buffer, sector_size);
        buffer += sector_size;
        ++sector;
        --count;
    }
    return 0;
}

uint32_t DFS_WriteSector(uint8_t unit, uint8_t *buffer, uint32_t sector, uint32_t count)
{
    while (count > 0) {
        vm_params[0] = scratch_buffer_off;
        vm_params[1] = sector_size;
        vm_params[2] = sector;
        CopyAndSwapBytes(scratch_buffer, buffer, sector_size);
        if (_syscall(NULL, SYS_mboxcmd, vm_mbox_off, SD_WRITE_CMD | (vm_params_off << 8)) != 0)
            return -1;
        buffer += sector_size;
        ++sector;
        --count;
    }
    return 0;
}
